'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

exports.setCacheValue = setCacheValue;
exports.computeValueAndDependencies = computeValueAndDependencies;
exports.createMemoizedGetter = createMemoizedGetter;
exports.createCopier = createCopier;
exports.groupShapeElements = groupShapeElements;
exports.default = createState;
/**
 * Set cache value of a given key on a given state
 *
 * @param state
 * @param cacheKey
 * @param value
 */
function setCacheValue(state, cacheKey, value) {
  Object.defineProperty(state, cacheKey, { enumerable: false, value: value });

  // decrement computations left
  state.__computationsLeft__--;

  // if there's no more computations left
  // delete previous state to decrease memory usage
  if (state.__computationsLeft__ === 0) delete state.__previous__;
}

/**
 * Compute a getter value and discover its dependencies
 *
 * @param state
 * @param getters
 * @param computedCacheKey
 * @param fn
 * @returns {*}
 */
function computeValueAndDependencies(state, getters, computedCacheKey, fn) {
  var updatedBy = {};

  // Create a proxy object to discover a getter dependencies
  var discoveryProxy = Object.keys(state).concat(getters).reduce(function (acc, it) {
    Object.defineProperty(acc, it, {
      get: function get() {
        updatedBy[it] = true;
        return state[it];
      }
    });
    return acc;
  }, {});

  setCacheValue(state, computedCacheKey, fn.call(discoveryProxy, discoveryProxy));

  fn.__updatedBy__ = Object.keys(updatedBy);

  return state[computedCacheKey];
}

/**
 * Create a memoized getter
 *
 * @param name
 * @param getters
 * @param fn
 * @returns {Function}
 */
function createMemoizedGetter(name, getters, fn) {
  var cacheKey = '@@' + name;
  var gettersKeys = Object.keys(getters);

  return function () {
    var _this = this;

    // If dependencies hasn't been discovered yet
    if (!fn.hasOwnProperty('__updatedBy__')) return computeValueAndDependencies(this, gettersKeys, cacheKey, fn);

    var didNotComputeYet = !this.hasOwnProperty(cacheKey);

    var shouldCompute = didNotComputeYet && (
    // has no previous state
    !this.hasOwnProperty('__previous__')

    // has no dependencies
    || fn.__updatedBy__.length === 0

    // some dependency has been updated
    || fn.__updatedBy__.some(function (key) {
      return _this[key] !== _this['__previous__'][key];
    }));

    if (shouldCompute) setCacheValue(this, cacheKey, fn.call(this, this));else if (didNotComputeYet) setCacheValue(this, cacheKey, this['__previous__'][name]);

    return this[cacheKey];
  };
}

/**
 * Create copier function
 *
 * @param numberOfGetters
 * @returns {Function}
 */
function createCopier(numberOfGetters) {

  return function () {
    var newValues = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

    var newState = Object.assign({}, this, newValues);

    Object.defineProperties(newState, {
      __previous__: {
        enumerable: false,
        writable: false,
        configurable: true,
        value: this
      },
      __computationsLeft__: {
        enumerable: false,
        writable: true,
        value: numberOfGetters
      }
    });

    newState.__proto__ = this.__proto__;

    return newState;
  };
}

/**
 * Group shape elements into initialState and getters
 *
 * @param shape
 * @returns {*}
 */
function groupShapeElements(shape) {
  return Object.keys(shape).reduce(function (acc, it) {
    switch (_typeof(shape[it])) {
      case 'function':
        acc.getters[it] = shape[it];
        break;

      default:
        acc.initialState[it] = shape[it];
        break;
    }

    return acc;
  }, { initialState: {}, getters: {} });
}

/**
 * Create a state container for a given shape
 *
 * @param shape
 */
function createState(shape) {
  var _groupShapeElements = groupShapeElements(shape),
      initialState = _groupShapeElements.initialState,
      getters = _groupShapeElements.getters;

  var numberOfGetters = Object.keys(getters).length;

  var proto = Object.keys(getters).reduce(function (acc, it) {
    Object.defineProperty(acc, it, {
      enumerable: true,
      get: createMemoizedGetter(it, getters, getters[it])
    });

    return acc;
  }, {});

  Object.defineProperty(proto, 'copy', {
    enumerable: false,
    writable: false,
    value: createCopier(numberOfGetters)
  });

  Object.defineProperty(initialState, '__computationsLeft__', {
    enumerable: false,
    writable: true,
    value: numberOfGetters
  });

  initialState.__proto__ = proto;

  return initialState;
}