'use strict';

var _chai = require('chai');

var _createState = require('../src/createState');

var _createState2 = _interopRequireDefault(_createState);

var _sinon = require('sinon');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

describe('createState', function () {

  it('should compute getters only when required fields change', function () {
    var deltaSpy = (0, _sinon.spy)();
    var echoSpy = (0, _sinon.spy)();

    var first = (0, _createState2.default)({
      name: 'first',
      alpha: 1,
      bravo: 2,
      charlie: 3,
      delta: function delta(_ref) {
        var alpha = _ref.alpha,
            bravo = _ref.bravo;

        deltaSpy();
        return alpha + bravo;
      },
      echo: function echo(_ref2) {
        var delta = _ref2.delta,
            charlie = _ref2.charlie;

        echoSpy();
        return delta + charlie;
      }
    });

    var second = first.copy({ alpha: 5, name: 'second' });
    var third = second.copy({ charlie: 5, name: 'third' });

    (0, _chai.expect)(first.delta).equals(3);
    (0, _chai.expect)(first.echo).equals(6);
    (0, _chai.expect)(second.echo).equals(10);
    (0, _chai.expect)(third.echo).equals(12);
    (0, _chai.expect)(deltaSpy.callCount).equals(2);
    (0, _chai.expect)(echoSpy.callCount).equals(3);
  });

  it('should create state with shape values and getters', function () {
    var shape = {
      alpha: 1,
      bravo: 2,
      charlie: 3,

      echo: function echo() {
        return this.delta + this.charlie;
      },
      delta: function delta() {
        return this.alpha + this.bravo;
      }
    };

    var state = (0, _createState2.default)(shape);

    (0, _chai.expect)(state.alpha).to.equal(shape.alpha);
    (0, _chai.expect)(state.bravo).to.equal(shape.bravo);
    (0, _chai.expect)(state.charlie).to.equal(shape.charlie);
    (0, _chai.expect)(state.delta).to.equal(shape.alpha + shape.bravo);
    (0, _chai.expect)(state.echo).to.equal(shape.alpha + shape.bravo + shape.charlie);
  });

  it('should copy state and change values partially', function () {

    var firstState = (0, _createState2.default)({
      alpha: 1,
      bravo: 2,

      charlie: function charlie() {
        return this.alpha + this.bravo;
      }
    });
    var secondState = firstState.copy({ alpha: 2 });

    (0, _chai.expect)(firstState.alpha).to.equal(1);
    (0, _chai.expect)(firstState.bravo).to.equal(2);
    (0, _chai.expect)(firstState.charlie).to.equal(3);
    (0, _chai.expect)(secondState.alpha).to.equal(2);
    (0, _chai.expect)(secondState.bravo).to.equal(2);
    (0, _chai.expect)(secondState.charlie).to.equal(4);
  });

  it('should recalculate getter after dependency change', function () {
    var bravoCalls = 0;

    var shape = {
      alpha: 1,
      bravo: function bravo() {
        bravoCalls++;
        return this.alpha;
      }
    };

    var firstState = (0, _createState2.default)(shape);

    (0, _chai.expect)(firstState.bravo).to.equal(1);
    (0, _chai.expect)(bravoCalls).to.equal(1);

    var secondState = firstState.copy({ alpha: 2 });

    (0, _chai.expect)(secondState.bravo).to.equal(2);
    (0, _chai.expect)(secondState.bravo).to.equal(2);
    (0, _chai.expect)(bravoCalls).to.equal(2);
  });

  it('should not recalculate getter if dependency did not change', function () {
    var charlieCalls = 0;

    var shape = {
      alpha: 1,
      bravo: 2,
      charlie: function charlie(_ref3) {
        var alpha = _ref3.alpha;

        charlieCalls++;
        return alpha;
      }
    };

    var firstState = (0, _createState2.default)(shape);
    var secondState = firstState.copy({ bravo: 5 });
    var thirdState = secondState.copy({ bravo: 6, alpha: 1 });

    (0, _chai.expect)(firstState.charlie).to.equal(shape.alpha);
    (0, _chai.expect)(secondState.charlie).to.equal(shape.alpha);
    (0, _chai.expect)(thirdState.charlie).to.equal(shape.alpha);
    (0, _chai.expect)(charlieCalls).to.equal(1);
  });

  it('should cache getter call after recalculation', function () {
    var bravoCalls = 0;

    var shape = {
      alpha: 1,
      bravo: function bravo() {
        bravoCalls++;
        return this.alpha;
      }
    };

    var firstState = (0, _createState2.default)(shape);

    (0, _chai.expect)(firstState.bravo).to.equal(shape.alpha);
    (0, _chai.expect)(firstState.bravo).to.equal(shape.alpha);
    (0, _chai.expect)(bravoCalls).to.equal(1);

    var secondState = firstState.copy({ alpha: 2 });

    (0, _chai.expect)(secondState.bravo).to.equal(2);
    (0, _chai.expect)(secondState.bravo).to.equal(2);
    (0, _chai.expect)(bravoCalls).to.equal(2);
  });
});